import sinon from 'sinon';

const getArrayFromArrayLikeObject = (args) => Array.prototype.slice.call(args);

const isMethod = (object) => (propName) =>
  !Object.getOwnPropertyDescriptor(object, propName).get &&
  typeof object[propName] === "function" &&
  !(propName === "constructor");

const applyToEachFunctionKeyInObject = (appliedFunction, object) =>
  Object.getOwnPropertyNames(object).filter(isMethod(object)).forEach(appliedFunction);

const applyToEachFunctionKeyInPrototypeChain = (appliedFunction, object) => {
  if (object) {
    applyToEachFunctionKeyInObject(appliedFunction, object);
    applyToEachFunctionKeyInPrototypeChain(appliedFunction, Object.getPrototypeOf(object));
  }
};

const getInstanceIndexWithValidation = (index, numInstances) => {
  const instanceIndex = index || 0;

  if (typeof index === "undefined") {
    if (numInstances > 1) {
      throw new Error(
        "Tried to access only instance of StubConstructor, " +
          `but there were ${numInstances} instances.`
      );
    }
  }
  if (numInstances <= instanceIndex) {
    throw new Error(
      `Tried to access StubConstructor instance ${instanceIndex}, ` +
        `but there were only ${numInstances} instances.`
    );
  }
  return instanceIndex;
};

var getStubOrSpyConstructor = /* tree-shaking no-side-effects-when-called */ (getConstructorProperties) =>
  (Target) => {
    const constructorProps = getConstructorProperties(Target);
    let init;

    function StubOrSpyConstructor() {
      constructorProps.SourceConstructor.apply(this, arguments);
      StubOrSpyConstructor.args.push(getArrayFromArrayLikeObject(arguments));
      if (this instanceof StubOrSpyConstructor) {
        StubOrSpyConstructor.instances.push(this);

        Target &&
          applyToEachFunctionKeyInPrototypeChain(
            constructorProps.processMethodOfInstance(this),
            constructorProps.getInstanceMethodNameSource(this)
          );
        init && init(this);
      } else {
        StubOrSpyConstructor.instances.push(null);
      }
    }

    StubOrSpyConstructor.prototype = constructorProps.SourceConstructor.prototype;

    StubOrSpyConstructor.withInit = function (onInit) {
      init = onInit;
      return this;
    };

    StubOrSpyConstructor.instances = [];

    StubOrSpyConstructor.getInstance = (index) => {
      const validatedIndex = getInstanceIndexWithValidation(
        index,
        StubOrSpyConstructor.instances.length
      );
      return StubOrSpyConstructor.instances[validatedIndex];
    };

    StubOrSpyConstructor.args = [];

    Target &&
      applyToEachFunctionKeyInObject(
        constructorProps.processMethodOfConstructor(StubOrSpyConstructor),
        Target
      );
    return StubOrSpyConstructor;
  };

const setMethodToStub = (object) => (methodName) => {
  if (!(methodName in object)) {
    object[methodName] = sinon.stub();
  }
};

const spyOnMethod = (object) => (methodName) => {
  if (
    !(methodName in Object.prototype) &&
    !(object[methodName] && object[methodName].isSinonProxy)
  ) {
    sinon.spy(object, methodName);
  }
};

const copyAndSpyOnMethod = (object, source) => (methodName) => {
  if (source[methodName].isSinonProxy) {
    object[methodName] = source[methodName];
  } else {
    object[methodName] = sinon.spy(source[methodName]);
  }
};

const getStubConstructorProperties = (Target) => ({
  SourceConstructor: function () {},
  processMethodOfInstance: setMethodToStub,
  getInstanceMethodNameSource: () => Target.prototype,
  processMethodOfConstructor: (TheConstructor) => setMethodToStub(TheConstructor),
});

const getSpyConstructorProperties = (Target) => ({
  SourceConstructor: Target,
  processMethodOfInstance: spyOnMethod,
  getInstanceMethodNameSource: (instance) => instance,
  processMethodOfConstructor: (TheConstructor) => copyAndSpyOnMethod(TheConstructor, Target),
});

const getStubConstructor = /* @__PURE__ */ getStubOrSpyConstructor(
  getStubConstructorProperties
);

const getSpyConstructor = /* @__PURE__ */ getStubOrSpyConstructor(
  getSpyConstructorProperties
);

export { getSpyConstructor, getStubConstructor };
